using System.Collections;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;

public class TileManager : MonoBehaviour
{
    public int gridLength = 31;
    public int gridWidth = 31;
    public Tile[,] tilesGrid;
    public Transform tilesTransform;
    public GameObject tilePrefab;
    public GameObject shrineTilePrefab;
    public GameObject endTilePrefab;

    public SpecialTileSpawner specialTileSpawner;

    public float perlinNoise;
    public float refinement;
    public float multiplier;

    [HideInInspector]
    public ShrineSpawn initialShrine;

    // Start is called before the first frame update
    void Start()
    {
        GenerateGrid();
    }

    // Update is called once per frame
    void Update()
    {
    }

    private void GenerateGrid()
    {
        tilesGrid = new Tile[gridLength, gridWidth];

        for (int x = 0; x < gridLength; x++)
        {
            for (int z = 0; z < gridWidth; z++)
            {
                perlinNoise = Mathf.PerlinNoise(x * refinement, z * refinement);
                tilesGrid[x, z] = CreateTile(x, z);
            }
        }

        CreateSpawnCluster();
        CreateEndCluster();
        SpawnSpecialTiles();
    }

    private void CreateSpawnCluster()
    {
        Tile shrineTile = SwapTile(shrineTilePrefab, gridLength / 2, 2);
        initialShrine = shrineTile.transform.GetComponentInChildren<ShrineSpawn>();
        LiftTile(shrineTile, true);
        LiftTiles(GetSurroundingTiles(shrineTile, false, 2), true);
        LiftRandomSpawnTile(12);
    }

    private void CreateEndCluster()
    {
        Tile winTile = SwapTile(endTilePrefab, gridLength / 2, gridWidth - 3);
        LiftTile(winTile, true);
        LiftTiles(GetSurroundingTiles(winTile, false), true);
    }

    private void SpawnSpecialTiles()
    {
        for (int i = 0; i < specialTileSpawner.tilesToSpawn; i++)
        {
            for (int j = 0; j < specialTileSpawner.maxIndependantAttempts; j++)
            {
                int x = Random.Range(0, tilesGrid.GetLength(0));
                int y = Random.Range(0, tilesGrid.GetLength(1));
                Tile tempTile = tilesGrid[x, y];
                if(!SpecialInRadius(tempTile))
                {
                    SwapTile(specialTileSpawner.SelectTile().gameObject, x, y);
                    j = specialTileSpawner.maxIndependantAttempts;
                }
            }
        }
    }

    private Tile SwapTile(GameObject tilePrefab, int x, int y)
    {
        Tile tempTile = tilesGrid[x, y];
        Tile tile = Instantiate(tilePrefab, tempTile.transform.position, Quaternion.identity, tilesTransform).GetComponent<Tile>();
        Destroy(tempTile.gameObject);
        tilesGrid[x, y] = tile;
        tilesGrid[x, y].Initialize(x, y);
        return tilesGrid[x, y];
    }

    private bool SpecialInRadius(Tile tile)
    {
        bool hasSpecial = tile.isSpecial;
        Tile[] tiles = GetSurroundingTiles(tile, false, 3);
        foreach (Tile t in tiles)
        {
            if (t.isSpecial) return true;
        }
        return hasSpecial;
    }

    private Tile CreateTile(int x, int z)
    {
        Tile tile = Instantiate(tilePrefab, tilesTransform.position + new Vector3(3 * x, -3 + (perlinNoise * multiplier), 3 * z), Quaternion.identity, tilesTransform).GetComponent<Tile>();
        tile.Initialize(x, z);
        return tile;
    }

    public void LiftTile(Tile tile, bool instant = false)
    {
        tile.Lift(instant);
    }

    public void LiftTiles(Tile[] tiles, bool instant = false)
    {
        foreach (Tile tile in tiles)
        {
            tile.Lift(instant);
        }
    }

    public void LiftRandomTiles(Tile originalTile, int amount)
    {
        for (int i = 0; i < amount; i++)
        {
            LiftRandomTile(originalTile);
        }
    }

    public void LiftRandomTile(Tile originalTile)
    {
        Tile[] tiles = GetInvertedSurroundingTiles(originalTile);

        if (tiles.Length <= 0) return;

        tiles[Random.Range(0, tiles.Length)].Lift();
    }

    public void LiftRandomSpawnTile(int amount = 1)
    {
        for (int i = 0; i < amount; i++)
        {
            LiftRandomTile(initialShrine.GetComponentInParent<Tile>());
        }
    }

    public Tile[] GetInvertedSurroundingTiles(Tile originalTile, int radius = 1)
    {
        Tile[] tiles = GetSurroundingTiles(GetAllSurroundingTiles(originalTile), !originalTile.lifted, radius);
        return tiles;
    }

    public Tile[] GetSurroundingTiles(Tile[] tiles, bool lifted, int radius = 1)
    {
        List<Tile> tilesList = new List<Tile>();

        foreach (Tile tile in tiles)
        {
            foreach (Tile tileGrid in GetSurroundingTiles(tile, lifted, radius))
            {
                if (Mathf.Abs(tileGrid.x - tile.x) <= radius && Mathf.Abs(tileGrid.z - tile.z) <= radius && tileGrid.lifted == lifted)
                {
                    if (!tilesList.Contains(tileGrid) && !tiles.Contains(tileGrid))
                    {
                        tilesList.Add(tileGrid);
                    }
                }
            }
        }

        return tilesList.ToArray();
    }

    public Tile[] GetSurroundingTiles(Tile tile, bool lifted, int radius = 1)
    {
        List<Tile> tiles = new List<Tile>();

        foreach (Tile tileGrid in tilesGrid)
        {
            if (Mathf.Abs(tileGrid.x - tile.x) <= radius && Mathf.Abs(tileGrid.z - tile.z) <= radius && tileGrid.lifted == lifted)
            {
                tiles.Add(tileGrid);
            }
        }

        return tiles.ToArray();
    }

    public Tile[] GetAllSurroundingTiles(Tile tile)
    {
        List<Tile> tiles = new List<Tile>();
        tiles.Add(tile);

        while (true)
        {
            Tile[] tempTiles = GetSurroundingTiles(tiles.ToArray(), tile.lifted);

            if (tempTiles.Length == 0) 
            {
                break;
            }

            foreach (Tile item in tempTiles)
            {
                tiles.Add(item);
            }
        }

        return tiles.ToArray();
    }

    [System.Serializable]
    public class SpecialTileSpawner
    {
        public int tilesToSpawn = 8;
        public int maxIndependantAttempts = 20;
        public SpecialTileSelector[] tileSelectors;

        public Tile SelectTile()
        {
            int rnd = Random.Range(0, TotalWeight());
            int cur = 0;
            for (int i = 0; i < tileSelectors.Length; i++)
            {
                if (rnd < cur + tileSelectors[i].weighting && rnd >= cur)
                {
                    return tileSelectors[i].tile;
                }
                cur += tileSelectors[i].weighting;
            }

            return tileSelectors[0].tile;
        }

        private int TotalWeight()
        {
            int total = 0;
            foreach (SpecialTileSelector sts in tileSelectors)
            {
                total += sts.weighting;
            }
            return total;
        }
    }

    [System.Serializable]
    public class SpecialTileSelector
    {
        public Tile tile;
        public int weighting = 1;
    }
}
